class RouteManager(object):
    def __init__(self, config):
        self.route_mapper = config.get_routes_mapper()
        self.introspector = config.introspector

    def get_routes(self):
        for route in self.route_mapper.get_routes():
            route.pattern = route.pattern.replace('{optional_slash:/?}', '')
            yield route

    def get_route(self, route_name):
        for route in self.get_routes():
            if route.name == route_name:
                return route

    def get_view_name(self, route_name):
        view_func = self.get_view_func(route_name)

        if view_func:
            return self.get_function_name(view_func)
        else:
            return None

    def get_view_func(self, route_name):
        route_intr = self.introspector.get('routes', route_name)

        if not route_intr:
            return None

        for related in self.introspector.related(route_intr):
            if related.category_name == 'views':
                if related['attr']:
                    return getattr(related['callable'], related['attr'])
                else:
                    return related['callable']

    def get_function_name(self, obj):
        # obj could be a method or a function
        if hasattr(obj, '__qualname__'):  # a method in Python 3
            return '%s.%s' % (
                obj.__module__, obj.__qualname__)
        if hasattr(obj, 'im_class'):  # a method in Python 2
            return '%s.%s.%s' % (
                obj.__module__, obj.im_class.__name__, obj.__name__)
        else:  # a function
            return '%s.%s' % (
                obj.__module__, obj.__name__)



from pyramid import testing
from pyramid.httpexceptions import HTTPNotFound
from pytest import raises
from webtest import TestApp


def _make_config():
    config = testing.setUp()
    config.include('tomb_routes')
    return config


def _make_app(config=None):
    if config is None:
        config = _make_config()

    app = config.make_wsgi_app()
    return TestApp(app)


def test_routes_index():
    from tests.simple_app import my_view

    config = _make_config()
    config.add_simple_route('/path/to/view', my_view, renderer='json')

    config.include('tomb_reflect', route_prefix='/api_info')
    config.tomb_reflect_add_views()

    response = _make_app(config).get('/api_info/routes', status=200)

    assert response.content_type == 'application/json'

    route_info = (response.json.get('my_view')             # New tomb_routes
                  or response.json.get('/path/to/view'))   # Old tomb_routes
    assert route_info['route_pattern'] == '/path/to/view'

    route_info = response.json['tomb_reflect.routes#index']
    assert route_info['route_pattern'] == 'api_info/routes'

    route_info = response.json['tomb_reflect.routes#show']
    assert route_info['route_pattern'] == 'api_info/routes/{route_name}'


def test_routes_show_imperative():
    from tests.simple_app import my_view

    config = _make_config()
    config.add_simple_route('/path/to/view', my_view, renderer='json')

    config.include('tomb_reflect', route_prefix='/api_info')
    config.tomb_reflect_add_views()

    response = _make_app(config).get('/api_info/routes', status=200)
    assert response.content_type == 'application/json'

    for route_name in ('my_view', '/path/to/view'):
        if route_name in response.json.keys():
            break

    response = _make_app(config).get(
        '/api_info/routes/{route_name}'.format(route_name=route_name),
        status=200)

    assert response.content_type == 'application/json'
    assert response.json['route_pattern'] == '/path/to/view'
    assert response.json['route_name'] == 'my_view'
    assert response.json['view'] == 'tests.simple_app.my_view'


def test_routes_show_declarative():
    config = _make_config()
    config.scan('tests.simple_app')

    config.include('tomb_reflect', route_prefix='/api_info')
    config.tomb_reflect_add_views()

    response = _make_app(config).get(
        '/api_info/routes/decorated_view',
        status=200)

    assert response.content_type == 'application/json'
    assert response.json['route_pattern'] == '/path/to/decorated/view/func'
    assert response.json['route_name'] == 'decorated_view'
    assert response.json['view'] == 'tests.simple_app.decorated_view'


def test_routes_show_class_method():
    config = _make_config()
    config.scan('tests.simple_app')

    config.include('tomb_reflect', route_prefix='/api_info')
    config.tomb_reflect_add_views()

    response = _make_app(config).get(
        '/api_info/routes/MyViewsClass.imperative_view',
        status=200)

    assert response.content_type == 'application/json'
    assert response.json['route_pattern'] == '/path/to/decorated/view/method'
    assert response.json['route_name'] == 'MyViewsClass.imperative_view'
    assert 'imperative_view' in response.json['view']


def test_routes_show_route_not_found():
    config = _make_config()
    config.include('tomb_reflect', route_prefix='/api_info')
    config.tomb_reflect_add_views()

    with raises(HTTPNotFound):
        _make_app(config).get(
            '/api_info/routes/XYZ_route_not_found_XYZ',
            status=404)


def test_routes_show_route_with_no_view():
    route_name = 'route_with_no_view'
    path = '/route_with_no_view'
    config = _make_config()
    config.add_route(route_name, path)
    config.include('tomb_reflect', route_prefix='/api_info')
    config.tomb_reflect_add_views()

    response = _make_app(config).get(
        '/api_info/routes/route_with_no_view',
        status=200)

    assert response.content_type == 'application/json'
    assert response.json['route_pattern'] == '/route_with_no_view'
    assert response.json['route_name'] == 'route_with_no_view'
    assert response.json['view'] is None



from pyramid.httpexceptions import HTTPNotFound


class Routes(object):
    def __init__(self, request):
        self.request = request
        self.route_manager = request.registry.route_manager

    def index(self):
        return dict(
            (route.name, {'route_pattern': route.pattern})
            for route in self.route_manager.get_routes()
        )

    def show(self):
        """This returns info about a route."""

        route_name = self.request.matchdict['route_name']

        route = self.route_manager.get_route(route_name)
        if not route:
            raise HTTPNotFound(
                'The route with name %r could not be found' % route_name)
        view_name = self.route_manager.get_view_name(route_name)

        return {
            'route_name': route.name,
            'route_pattern': route.pattern,
            'view': view_name,
        }



"""
tomb_reflect

Make your API introspectable
"""

from .routemanager import RouteManager


URL_PATTERNS_TO_ACTIONS = {
    '/routes':               'index',
    '/routes/{route_name}':  'show',
}
ACTIONS = URL_PATTERNS_TO_ACTIONS.values()


def route_name_for_action(action):
    return 'tomb_reflect.routes#' + action


def tomb_reflect_add_views(config, *args, **kwargs):
    if not kwargs.get('renderer'):
        kwargs['renderer'] = 'json'

    for action in ACTIONS:
        config.add_view(
            'tomb_reflect.views.Routes', attr=action,
            route_name=route_name_for_action(action),
            *args, **kwargs)


def includeme(config):
    config.registry.route_manager = RouteManager(config)

    for url_pattern, action in URL_PATTERNS_TO_ACTIONS.items():
        route_name = route_name_for_action(action)
        config.add_route(name=route_name, pattern=url_pattern)

    # Client can do: config.tomb_reflect_add_views()
    # They can also add parameters for auth, etc.
    config.add_directive('tomb_reflect_add_views', tomb_reflect_add_views)



from tomb_routes import simple_route


def my_view(request):
    return {'foo': 'bar'}


@simple_route('/path/to/decorated/view/func', renderer='json')
def decorated_view(request):
    return {'foo': 'bar'}


@simple_route('/matchdict/{name}/{number}', renderer='json')
def matchdict_view(request, name, number):
    return {'foo': name, 'bar': number}


class MyViewsClass(object):
    def __init__(self, request):
        self.request = request

    @simple_route('/path/to/decorated/view/method', renderer='json')
    def imperative_view(self):
        return {'foo': 'bar'}

    @simple_route('/matchdict/{name}/{number}', renderer='json')
    def matchdict_view(self, name, number):
        return {'foo': name, 'bar': number}
